﻿Public Class BubbleCrusher
    Enum GameType As Byte
        SuperShift
        Shift
        Continuous
        Standard
        MaxGame
    End Enum

    Private GameMatrix(,) As Byte
    Private HighlightMatrix(,) As Byte
    Private myGame As GameType
    Private myColorCount As Byte
    Private myColors() As Color = {Color.Red, Color.Blue, Color.Green, Color.Gold, Color.Violet, Color.Orange, Color.Black, Color.Silver}
    Private myColorsFiles() As String = {"Red", "Blue", "Green", "Yellow", "Violet", "Orange", "Dark", "Gray"}
    Private myAvailableSkins() As String = {"Bubble", "SmileyBubble", "Atom", "Schlapa.Net Sun", "Star BubblePlanet"}
    Private CurrentSkin As Byte = 0
    Private myButtonImage() As Bitmap
    Private myFieldSize As Size
    Private myBubbleSize As Size

    Private myReady As Boolean
    Private myScore As Integer

    Private tempGraphics As Graphics
    Private RandomNumber As New Random()

    Public Event Redraw()

    Public ReadOnly Property AvailableSkins() As String()
        Get
            Return myAvailableSkins
        End Get
    End Property

    Public ReadOnly Property Matrix() As Byte(,)
        Get
            Return GameMatrix
        End Get
    End Property

    Public ReadOnly Property FieldSize() As Size
        Get
            Return myFieldSize
        End Get
    End Property

    Public ReadOnly Property BubbleSize() As Size
        Get
            Return myBubbleSize
        End Get
    End Property

    Public ReadOnly Property LastGame() As GameType
        Get
            Return myGame
        End Get
    End Property

    Public ReadOnly Property isReady() As Boolean
        Get
            Return myReady
        End Get
    End Property

    Public Function GetCrushScore(ByVal BubbleCount As Integer) As Integer
        Return BubbleCount * (BubbleCount - 1)
    End Function

    Sub New(ByVal Game As GameType, ByVal ColorCount As Byte, ByVal FieldSize As Size, ByVal BubbleSize As Size, ByVal Skin As Byte)
        myReady = False
        If ColorCount = 0 Then Throw New Exception("ColorCount must be greater than zero!")
        If Not [Enum].IsDefined(GetType(GameType), Game) Then Throw New Exception("The game you specified does not exist!")
        If FieldSize.Width = 0 Then Throw New Exception("FieldSize.Width must not be zero")
        If FieldSize.Height = 0 Then Throw New Exception("FieldSize.Height must not be zero")
        If ColorCount > myColors.Count Then Throw New Exception("More than " & (myColors.Count) & " different Colored buttons are not allowed!")

        myGame = Game
        myColorCount = ColorCount
        myFieldSize = FieldSize
        myBubbleSize = BubbleSize
        CurrentSkin = Skin

        AppPath = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase)
        If AppPath.StartsWith("file:\") Then
            AppPath = AppPath.Substring(6)
        End If

        Call SetUpColors()
        Call SetUpGameField()
        myReady = True
    End Sub

    Public Function GetScore() As Integer
        Return myScore
    End Function

    'Public Function GetButtonColor(ByVal FieldValue As Byte) As Color
    '    Return myColors(FieldValue - 1)
    'End Function

    Public Function Crush(ByVal CrushStart As Point, ByVal CrushList As List(Of Point))
        If CrushList.Count <= 1 Then Return False
        If Not CrushList.Contains(CrushStart) Then Return False

        For Each ToCrush As Point In CrushList
            GameMatrix(ToCrush.X, ToCrush.Y) = 0
        Next

        RaiseEvent Redraw()
        Threading.Thread.Sleep(100)

        If myGame = GameType.SuperShift Then
            Call FallDown()

            Threading.Thread.Sleep(150)
            RaiseEvent Redraw()

            Call GoRight()

            Threading.Thread.Sleep(150)

            If AddNew() Then
                RaiseEvent Redraw()

                Threading.Thread.Sleep(150)
                RaiseEvent Redraw()

                Call GoRight()
            End If
        ElseIf myGame = GameType.Shift Then
            Call FallDown()

            Threading.Thread.Sleep(150)
            RaiseEvent Redraw()

            Call GoRight()

            Threading.Thread.Sleep(150)
        ElseIf myGame = GameType.Continuous Then
            Call FallDown()

            Threading.Thread.Sleep(150)
            RaiseEvent Redraw()

            Call GoRightOnEmpty()

            Threading.Thread.Sleep(150)
            RaiseEvent Redraw()

            If AddNew() Then
                RaiseEvent Redraw()

                Threading.Thread.Sleep(150)
                RaiseEvent Redraw()

                Call GoRightOnEmpty()

                'Threading.Thread.Sleep(150)
                'RaiseEvent Redraw()

                Call AddNew()
                Threading.Thread.Sleep(150)
            End If
        ElseIf myGame = GameType.Standard Then
            Call FallDown()

            Threading.Thread.Sleep(150)
            RaiseEvent Redraw()

            Call GoRightOnEmpty()

            Threading.Thread.Sleep(150)
        End If

        RaiseEvent Redraw()
        myScore += GetCrushScore(CrushList.Count)
        Return True
    End Function

    Private Sub FallDown()
        Dim FirstEmpty As Point

        For X = 0 To myFieldSize.Width - 1
            FirstEmpty = Nothing
            For Y = myFieldSize.Height - 1 To 0 Step -1
                If GameMatrix(X, Y) = 0 And FirstEmpty = Nothing Then
                    FirstEmpty = New Point(X, Y)
                ElseIf GameMatrix(X, Y) > 0 And Not FirstEmpty = Nothing Then
                    GameMatrix(FirstEmpty.X, FirstEmpty.Y) = GameMatrix(X, Y)
                    GameMatrix(X, Y) = 0
                    FirstEmpty.Y -= 1
                End If
            Next
        Next
    End Sub

    Private Sub GoRight()
        Dim FirstEmpty As Point

        For Y = 0 To myFieldSize.Height - 1
            FirstEmpty = Nothing
            For X = myFieldSize.Width - 1 To 0 Step -1
                If GameMatrix(X, Y) = 0 And FirstEmpty = Nothing Then
                    FirstEmpty = New Point(X, Y)
                ElseIf GameMatrix(X, Y) > 0 And Not FirstEmpty = Nothing Then
                    GameMatrix(FirstEmpty.X, FirstEmpty.Y) = GameMatrix(X, Y)
                    GameMatrix(X, Y) = 0
                    FirstEmpty.X -= 1
                End If
            Next
        Next
    End Sub

    Private Sub GoRightOnEmpty()
        For X = myFieldSize.Width - 1 To 1 Step -1
            If GameMatrix(X, myFieldSize.Height - 1) = 0 Then
                For YPos = 0 To myFieldSize.Height - 1
                    For XPos = X To 1 Step -1
                        GameMatrix(XPos, YPos) = GameMatrix(XPos - 1, YPos)
                    Next
                    GameMatrix(0, YPos) = 0
                Next
            End If
        Next
    End Sub

    Private Function AddNew()
        Dim EmptyCount As Integer = 0

        For X = 0 To myFieldSize.Width - 1
            If GameMatrix(X, myFieldSize.Height - 1) > 0 Then
                Exit For
            Else
                EmptyCount += 1
            End If
        Next

        If EmptyCount > 0 Then
            For X = 0 To EmptyCount - 1
                For Y = myFieldSize.Height - 1 To RandomNumber.Next(0, myFieldSize.Height - 1) Step -1
                    GameMatrix(X, Y) = RandomNumber.Next(1, myColorCount + 1)
                Next
            Next
            Return True
        Else
            Return False
        End If
    End Function

    Public Function CheckGameDone() As Boolean
        Dim tempColor As Byte
        For X = 0 To myFieldSize.Width - 1
            For Y = 0 To myFieldSize.Height - 1
                tempColor = GameMatrix(X, Y)
                If tempColor > 0 Then
                    If Y + 1 < myFieldSize.Height Then
                        If GameMatrix(X, Y + 1) = tempColor Then
                            Return False
                        End If
                    End If
                    If X + 1 < myFieldSize.Width Then
                        If GameMatrix(X + 1, Y) = tempColor Then
                            Return False
                        End If
                    End If
                End If
            Next
        Next
        Return True
    End Function

    Public Function GetButtonImage(ByVal FieldValue As Byte) As Bitmap
        Return myButtonImage(FieldValue - 1)
    End Function

    Public Function GetHighlight(ByVal Position As Point) As List(Of Point)
        Dim tempColor As Byte = GameMatrix(Position.X, Position.Y)
        Dim tempPosition As Point
        Dim knownPos As New List(Of Point)

        knownPos.Add(Position)
        Dim CurrentPos As Integer = 0
        Do While CurrentPos < knownPos.Count
            tempPosition = knownPos.Item(CurrentPos)
            If tempPosition.Y - 1 >= 0 Then
                If GameMatrix(tempPosition.X, tempPosition.Y - 1) = tempColor And Not knownPos.Contains(New Point(tempPosition.X, tempPosition.Y - 1)) Then
                    knownPos.Add(New Point(tempPosition.X, tempPosition.Y - 1))
                End If
            End If
            If tempPosition.Y + 1 < myFieldSize.Height Then
                If GameMatrix(tempPosition.X, tempPosition.Y + 1) = tempColor And Not knownPos.Contains(New Point(tempPosition.X, tempPosition.Y + 1)) Then
                    knownPos.Add(New Point(tempPosition.X, tempPosition.Y + 1))
                End If
            End If

            If tempPosition.X - 1 >= 0 Then
                If GameMatrix(tempPosition.X - 1, tempPosition.Y) = tempColor And Not knownPos.Contains(New Point(tempPosition.X - 1, tempPosition.Y)) Then
                    knownPos.Add(New Point(tempPosition.X - 1, tempPosition.Y))
                End If
            End If
            If tempPosition.X + 1 < myFieldSize.Width Then
                If GameMatrix(tempPosition.X + 1, tempPosition.Y) = tempColor And Not knownPos.Contains(New Point(tempPosition.X + 1, tempPosition.Y)) Then
                    knownPos.Add(New Point(tempPosition.X + 1, tempPosition.Y))
                End If
            End If
            CurrentPos += 1
        Loop

        Return knownPos
    End Function

    Public Sub SetUpColors(ByVal newBubbleSize As Size)
        myBubbleSize = newBubbleSize
        Call SetUpColors()
    End Sub

    Public Sub SetUpColors(ByVal SkinID As Byte)
        CurrentSkin = SkinID
        Call SetUpColors()
    End Sub

    Public Sub SetUpColors()
        Dim tempName As String
        Dim tempBitmap As Bitmap
        Dim I As Integer = 0
        ReDim myButtonImage(myColorCount - 1)
        For I = 0 To myColorCount - 1
            tempName = "img\" & myAvailableSkins(CurrentSkin) & "\" & myAvailableSkins(CurrentSkin) & "-" & myColorsFiles(I) & ".jpg"
            myButtonImage(I) = New Bitmap(BubbleSize.Width, BubbleSize.Height)
            tempGraphics = Graphics.FromImage(myButtonImage(I))
            If IO.File.Exists(IO.Path.Combine(AppPath, tempName)) Then
                tempBitmap = New Bitmap(IO.Path.Combine(AppPath, tempName))
                tempGraphics.DrawImage(tempBitmap, New Rectangle(0, 0, BubbleSize.Width, BubbleSize.Height), New Rectangle(0, 0, tempBitmap.Width, tempBitmap.Height), GraphicsUnit.Pixel)
            Else
                tempGraphics.Clear(Color.White)
                tempGraphics.FillEllipse(New SolidBrush(myColors(I)), 2, 2, BubbleSize.Width - 4, BubbleSize.Height - 4)
            End If
        Next
        RaiseEvent Redraw()
    End Sub

    Private Sub SetUpGameField()
        ReDim GameMatrix(myFieldSize.Width - 1, myFieldSize.Height - 1)

        For X = 0 To GameMatrix.GetLength(0) - 1
            For Y = 0 To GameMatrix.GetLength(1) - 1
                GameMatrix(X, Y) = RandomNumber.Next(1, myColorCount + 1)
            Next
        Next
    End Sub
End Class
